# import asyncio
# from asyncwica import WicaStream

# async def simple_example():
#     """A simple example of how to use AsyncWicaStream. Run it in main by uncommenting it! """

#     wica_stream = WicaStream(base_url="http://student08/ca/streams", channels=["MMAC3:STR:2"])

#     async def run_stream():
#         await wica_stream.create()
#         async for message in wica_stream.subscribe():
#             print(message)

#     async def stop_stream():
#         await asyncio.sleep(10)
#         print(await wica_stream.destroy())

#     await asyncio.gather(run_stream(), stop_stream())

# async def example_using_with():
#     """ An example using the compound statement async with and another method to exit the event loop. Run it in main by uncommenting it!"""
#     async with WicaStream(base_url="http://student08/ca/streams", channels=["MMAC3:STR:2"]) as stream:
#         i:int = 0
#         async for message in stream.subscribe():
#             i+=1
#             print(message)
#             if i == 25:
#                 break

# async def multistream_example():
#     """ An example of how to run multiple streams at once using aiostream. Run it in main by uncommenting it! """
#     from aiostream import stream
#     streams = []
#     async def run_streams():
#         for _ in range(10):
#             wica_stream = WicaStream(base_url="http://student08/ca/streams", channels=["MMAC3:STR:2"])
#             streams.append(wica_stream)
#             await wica_stream.create()

#         print("Doing someting else before starting the stream...")
#         await asyncio.sleep(5)

#         subscribed_streams = []

#         for wica_stream in streams:
#             print(f"Subscribing to stream {wica_stream.id}")
#             subscribed_streams.append(wica_stream.subscribe())


#         combine = stream.merge(*subscribed_streams)
#         async with combine.stream() as streamer:
#             async for item in streamer:
#                 print(item)
#                 continue


#     async def stop_streams():
#         await asyncio.sleep(25)
#         for wica_stream in streams:
#             print(await wica_stream.destroy())


#     await asyncio.gather(run_streams(), stop_streams())


# async def main():
#     #await simple_example()
#     #await example_using_with()
#     #await multistream_example()
#     pass

# if __name__ == "__main__":
#     asyncio.run(main())



from dataclasses import dataclass
from datetime import datetime
from typing import List, Literal, Optional


@dataclass
class WicaStreamProperties:
    """Properties of a Wica stream.

    All intervals in milliseconds
    """

    qiet_mode: bool
    heartbeat_flux_nterval: int
    metadata_flux_interval: int
    monitored_value_flux_interval: int
    polled_value_flux_interval: int
    data_aquisition_mode: Literal["poll", "monitor", "poll-monitor", "poll-and-monitor"]
    polling_interval: int
    fields_of_interest: List[str]
    numeric_precision: int
    filter_type: Literal[
        "ALL_VALUE", "AVERAGER", "CHANGE_DETECTOR", "LAST_N", "ONE_IN_M", "RATE_LIMITER"
    ]
    filter_num_samples: int
    filter_num_samples_avg: int
    filter_cycle_length: int
    filter_sampling_interval: int
    filter_deadband: float


@dataclass
class WicaChannel:
    @dataclass
    class WicaChannelProperties:
        data_aquisition_mode: Optional[
            Literal["poll", "monitor", "poll-monitor", "poll-and-monitor"]
        ] = None
        polling_interval: Optional[int] = None
        fields_of_interest: List[str] = None
        filter_type: Optional[
            Literal[
                "ALL_VALUE",
                "AVERAGER",
                "CHANGE_DETECTOR",
                "LAST_N",
                "ONE_IN_M",
                "RATE_LIMITER",
            ]
        ] = None
        n: Optional[int] = None
        m: Optional[int] = None
        x: Optional[int] = None
        filter_sampling_interval: Optional[int] = None
        filter_deadband: Optional[float] = None

    name: str
    properties: WicaChannelProperties = None


@dataclass
class WicaMessage(object):
    stream_id: int = None
    event: Optional[Literal["heartbeat", "metadata", "value"]] = None
    value_type: Optional[Literal["polled", "monitored"]] = None
    data: Optional[dict] = None
    time: Optional[datetime] = None



import json
import logging
import re
from datetime import datetime
from typing import AsyncIterator, List

import httpx

from asyncwica.exceptions import (
    StreamNotCreatedException,
    StreamNotDeletedException,
    StreamSubscriptionException,
)
from asyncwica.sseclient import WicaSSEClient
from asyncwica.utils import WicaChannel, WicaMessage, WicaStreamProperties


class WicaStream(object):
    """An asynchronous version of the WicaStream object.

    An object of this class represents a single stream from an wica-http server.

    Usage:

    ```python
    >>> stream = wica_api.AsyncWicaStream(channels="EXAMPLE:CHANNEL:123")
    >>> async for message in stream.subscribe():
    >>>     print(message)
    ```

    Parameters:
        channels (List[str]|List[WicaChannel]): The channel or channels the stream should include. Optional parameters can be given per channel with the `WicaChannel` object.
                                                If only one channel is needed just encapsulate it.
        base_url (str): The endpoint the request should be sent to. Usually it has the form `SERVER:PORT/ca/streams`.
        stream_properties (WicaStreamProperties): The properties the stream should have. See `WicaStreamProperties` class for more info.

    """

    def __init__(
        self,
        channels: List[str] | List[WicaChannel],
        stream_properties: WicaStreamProperties | None = None,
        base_url: str = "127.0.0.1/ca/streams",
    ) -> None:
        self.id: int = None
        self.base_url = base_url
        self.channels = [
            WicaChannel(name=channel) if isinstance(channel, str) else channel
            for channel in channels
        ]
        self.stream_properties = stream_properties
        self._logger = logging.getLogger(self.__class__.__module__)
        self._logger.debug(
            f"Initialized AsyncWicaStream for channels {channels} with base_url {base_url}"
        )

    async def __aenter__(self):
        self.id = await self.create()
        return self

    async def __aexit__(self, _channel, _properties, _url):
        _ = await self.destroy()
        self.id = None
        return self

    async def create(self):
        """Create a stream from the wica-http server.

        This method can be used to pre-create streams before subscribing. It sends a POST request to the server and gets an ID.
        This ID is than written to the id field of the object and returned.

        Usage:

        ```python
        >>> for _ in range(10):
        >>>     wica_stream = AsyncWicaStream(base_url="http://student08/ca/streams", channels=["MMAC3:STR:2"])
        >>>     streams.append(wica_stream)
        >>>     await wica_stream.create()
        >>>
        >>> print("Doing someting else before starting the stream...")
        >>> await asyncio.sleep(5)
        >>>
        >>> subscribed_streams = []
        >>>
        >>> for wica_stream in streams:
        >>>     print(f"Subscribing to stream {wica_stream.id}")
        >>>     subscribed_streams.append(wica_stream.subscribe())
        ```

        See examples for more detail!


        Retruns:
            int: The ID of the created stream

        """
        channels_payload = []
        if self.channels:
            for channel in self.channels:
                if channel.properties:
                    channels_payload.append(
                        {"name": channel.name, "props": channel.properties.__dict__}
                    )
                else:
                    channels_payload.append({"name": channel.name})

        if self.stream_properties:
            payload = {"channels": channels_payload, "props": WicaStreamProperties}
        else:
            payload = {"channels": channels_payload}
        async with httpx.AsyncClient() as client:
            try:
                response = await client.post(url=self.base_url, json=payload)
                if response.is_success:
                    self.id = int(response.json())
                    self._logger.debug(
                        f"successfully started stream with id: {self.id}"
                    )
                    return self.id
                else:
                    response.raise_for_status()

            except httpx.RequestError as exc:
                self._logger.debug(
                    f"A RequestError occurred while creating stream {exc.request.url!r}."
                )
                raise StreamNotCreatedException(
                    f"A RequestError occurred while creating stream {exc.request.url!r}."
                )
            except httpx.HTTPStatusError as exc:
                self._logger.debug(
                    f"StatusError response {exc.response.status_code} while creating stream {exc.request.url!r}."
                )
                raise StreamNotCreatedException(
                    f"StatusError response while creating stream {exc.request.url!r}."
                )

    async def subscribe(self) -> AsyncIterator[WicaMessage]:
        """Subscribe to a stream.

        If the stream doesn't exists this function will call create() to create a new stream.


        Yields:
            AsyncIterator[WicaMessage]: That returns the messages sent by the server.
        """
        if not self.id:
            try:
                self._logger.debug("Trying to create stream")
                await self.create()
            except (StreamNotCreatedException):
                self._logger.debug("Couldn't create stream")

        async with httpx.AsyncClient() as client:
            # Compiling regex for segmenting the messages returned by the client
            regex_event = re.compile(r"(?<=\-)(\w+)$")
            regex_value_type = re.compile(r"(?<=\w\s)([a-z]+)")
            regex_time = re.compile(r"(.*)(?=\s\-\s\w)")
            wica_client = None
            try:
                # get the stream
                async with client.stream(
                    "GET", url=self.base_url + f"/{self.id}"
                ) as response:
                    if response.is_success:
                        wica_client = WicaSSEClient(response.aiter_bytes())
                        async for event in wica_client.events():
                            message = WicaMessage()
                            message.stream_id = int(event.id)
                            message.event = re.search(regex_event, event.event).group(0)
                            message.value_type = re.search(
                                regex_value_type, event.msg_info
                            ).group(0)
                            message.data = (
                                {"time": f"{event.data}"}
                                if message.event == "heartbeat"
                                else json.loads(event.data)
                            )
                            _time = re.search(regex_time, event.msg_info).group(0)
                            message.time = (
                                datetime.fromisoformat(_time) if _time else None
                            )
                            yield message
                    else:
                        response.raise_for_status()

            except httpx.RequestError as exc:
                self._logger.debug(
                    f"A RequestError error occurred while subscribing to the stream {exc.request.url!r}."
                )
                raise StreamSubscriptionException(
                    f"A RequestError error occurred while subscribing to the stream  {exc.request.url!r}."
                )
            except httpx.HTTPStatusError as exc:
                self._logger.debug(
                    f"StatusError response {exc.response.status_code} while subscribing to the stream  {exc.request.url!r}."
                )
                raise StreamSubscriptionException(
                    f"A StatusError occurred while subscribing to the stream  {exc.request.url!r}."
                )
            finally:
                if wica_client:
                    await wica_client.close()

    async def destroy(self) -> httpx.Response:
        """Destroys the stream on the server.

        This method sends a delete request to the server and thereby destroys the stream.
        This is also the best way to stop a running subscription.

        Returns:
            Response: Returns the http response
        """
        async with httpx.AsyncClient() as client:
            try:
                if not self.id:
                    self._logger.debug(
                        "Stream was no deleted because no ID was provided!"
                    )
                    raise StreamNotDeletedException(
                        "Stream was not deleted because no ID was provided!"
                    )
                response = await client.delete(url=self.base_url + f"/{self.id}")
                if response.is_success:
                    return response
                else:
                    response.raise_for_status()
            except httpx.RequestError as exc:
                self._logger.debug(
                    f"A RequestError occurred while deleting stream  {exc.request.url!r}."
                )
                raise StreamNotDeletedException(
                    f"A RequestError occurred while deleting stream {exc.request.url!r}."
                )
            except httpx.HTTPStatusError as exc:
                self._logger.debug(
                    f"StatusError response {exc.response.status_code} while deleting stream  {exc.request.url!r}."
                )
                raise StreamNotDeletedException(
                    f"StatusError response {exc.response.status_code} while deleting stream  {exc.request.url!r}."
                )



from .exceptions import StreamNotCreatedException as StreamNotCreatedException
from .exceptions import StreamNotDeletedException as StreamNotDeletedException
from .exceptions import StreamSubscriptionException as StreamSubscriptionException
from .sseclient import WicaSSEClient as WicaSSEClient
from .stream import WicaStream as WicaStream
from .utils import WicaChannel as WicaChannel
from .utils import WicaMessage as WicaMessage
from .utils import WicaStreamProperties as WicaStreamProperties

__all__ = [
    "StreamNotCreatedException",
    "StreamSubscriptionException",
    "StreamNotDeletedException",
    "WicaSSEClient",
    "WicaStream",
    "WicaChannel",
    "WicaStreamProperties",
    "WicaMessage",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['asyncwica']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.23.0,<0.24.0']

setup_kwargs = {
    'name': 'asyncwica',
    'version': '1.2.2',
    'description': 'A simple async python API to access wica-http SSE.',
    'long_description': '# PyWica - Async Wica Python API\n[![pipeline status](https://git.psi.ch/proscan_data/py-wica/badges/async/pipeline.svg)](https://git.psi.ch/proscan_data/py-wica/-/commits/async)\n[![coverage report](https://git.psi.ch/proscan_data/py-wica/badges/async/coverage.svg)](https://git.psi.ch/proscan_data/py-wica/-/commits/async)\n\n#### Table of Contents\n- [Introduction](#introduction)\n- [Installation](#installation)\n- [Quick-start Guid](#quick-start-guide)\n- [Documentation](#documentation)\n- [Dependencies](#dependencies)\n- [Contribute](#contribute)\n- [Project Changes and Tagged Releases](#project-changes-and-tagged-releases)\n- [Developer Notes](#developer-notes)\n- [Contact](#contact)\n\n# Introduction\nThis project/package aims to provide a simple python interface to the wica-http server.\nCheck out the main branch to get the blocking version of the package\n\n# Installation\nInstall with pip\n```bash\npip install asyncwica\n```\n# Quick-start Guide\nHere are some simple examples to get you started:\n```python\nimport asyncio\nimport time\n\n\nasync def simple_example():\n    """A simple example of how to use AsyncWicaStream. Run it in main by uncommenting it! """\n\n    wica_stream = WicaStream(base_url="http://student08/ca/streams", channels=["MMAC3:STR:2"])\n\n    async def run_stream():\n        await wica_stream.create()\n        async for message in wica_stream.subscribe():\n            print(message)\n\n    async def stop_stream():\n        await asyncio.sleep(10)\n        print(await wica_stream.destroy())\n\n    await asyncio.gather(run_stream(), stop_stream())\n\nasync def example_using_with():\n    """ An example using the compound statement async with and another method to exit the event loop. Run it in main by uncommenting it!"""\n    async with WicaStream(base_url="http://student08/ca/streams", channels=["MMAC3:STR:2"]) as stream:\n        i:int = 0\n        async for message in stream.subscribe():\n            i+=1\n            print(message)\n            if i == 25:\n                break\n\nasync def multistream_example():\n    """ An example of how to run multiple streams at once using aiostream. Run it in main by uncommenting it! """\n    from aiostream import stream\n    streams = []\n    async def run_streams():\n        for _ in range(10):\n            wica_stream = WicaStream(base_url="http://student08/ca/streams", channels=["MMAC3:STR:2"])\n            streams.append(wica_stream)\n            await wica_stream.create()\n\n        print("Doing someting else before starting the stream...")\n        await asyncio.sleep(5)\n\n        subscribed_streams = []\n\n        for wica_stream in streams:\n            print(f"Subscribing to stream {wica_stream.id}")\n            subscribed_streams.append(wica_stream.subscribe())\n\n\n        combine = stream.merge(*subscribed_streams)\n        async with combine.stream() as streamer:\n            async for item in streamer:\n                print(item)\n                continue\n\n\n    async def stop_streams():\n        await asyncio.sleep(25)\n        for wica_stream in streams:\n            print(await wica_stream.destroy())\n\n\n    await asyncio.gather(run_streams(), stop_streams())\n\n\nasync def main():\n    #await simple_example()\n    #await example_using_with()\n    #await multistream_example()\n    pass\n\nif __name__ == "__main__":\n    asyncio.run(main())\n\n```\n\n# Documentation\nCurrent Features:\n* Custom Client to handle be able to extract last line of SSE with timestamp and message type.\n* Simple functions to create, delete and subscribe to streams\n* Fully Async (blocking versions available in main branch)\n\nCheck out the wiki for more info!\n\n# Dependencies\n* [httpx](https://github.com/encode/httpx/)\n\n# Contribute\nTo contribute, simply clone the project.\nYou can uses ``` pip -r requirements.txt ``` or the make file to set up the project.\n\n\n# Project Changes and Tagged Releases\n* See the Changelog file for further information\n* Project releases are available in pypi\n\n# Developer Notes\nCurrently None\n\n# Contact\nIf you have any questions pleas contract \'niklas.laufkoetter@psi.ch\'\n',
    'author': 'Niklas Laufkoetter',
    'author_email': 'niklas.laufkoetter@psi.ch',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://git.psi.ch/proscan_data/py-wica/-/tree/async',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
